// ObserverSwarm.m for Opinion formation model. Paul Johnson August 9, 1999
// Portions inherited from SFI's Heatbugs application
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.


#import <collections.h>
#import <objectbase.h>
#import <analysis.h>
#import <gui.h>
#import "ListWrapper.h"
#import "WrappedListGrid2d.h"
#import "MyArguments.h"
#import "ObserverSwarm.h"
#import "ModelSwarm.h"


//#import <dmalloc.h>

@implementation ObserverSwarm

// createBegin: here we set up the default observation parameters.

+ createBegin: aZone
{
  ObserverSwarm *obj;
  id <ProbeMap> probeMap;
  
  // Superclass createBegin to allocate ourselves.

  obj = [super createBegin: aZone];

  // Fill in the relevant parameters (only one, in this case).

  obj->displayFrequency = 1;

  // Also, build a customized probe map. Without a probe map, the default
  // is to show all variables and messages. Here we choose to
  // customize the appearance of the probe, give a nicer interface.

  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  // Add in a bunch of variables, one per simulation parameters

  [probeMap addProbe: [probeLibrary getProbeForVariable: "displayFrequency"
				    inClass: [self class]]];

//   [probeMap addProbe: [[probeLibrary getProbeForMessage: "graphBug:"
// 			     inClass: [self class]]
// 			setHideResult: YES]];

  // Now install our custom probeMap into the probeLibrary.

  [probeLibrary setProbeMap: probeMap For: [self class]];

  return obj;
}

// createEnd: create objects we know we'll need. In this case, none,
// but you might want to override this.

- createEnd
{
  return [super createEnd];
}

- _worldRasterDeath_ : caller
{
  [worldRaster drop];
  worldRaster = nil;
  return self;
}

- _unhappyGraphDeath_ : caller
{
  [opinionGraph drop];
  opinionGraph = nil;
  return self;
}

// Create the objects used in the display of the model. This code is
// fairly complicated because we build a fair number of widgets. It's
// also a good example of how to use the display code.

- buildObjects
{
  int i;

  [super buildObjects];
  
   modelSwarm = [ModelSwarm create: self];

  CREATE_ARCHIVED_PROBE_DISPLAY (modelSwarm);
  CREATE_ARCHIVED_PROBE_DISPLAY (self);
 

  [controlPanel setStateStopped];


  [modelSwarm buildObjects];
  [ModelSwarm setObserver: self];



  // Now get down to building our own display objects.

  // First, create a colormap: this is a global resource, the information
  // here is used by lots of different objects.

  colormap = [Colormap create: self];

  // Colours [0,64) are assigned to the range Red [0, 1), for heat display.

  for (i = 0; i < 64; i++)
    [colormap setColor: i ToRed: (double)i / 63.0 Green: 0 Blue: 0];

  // Colour 64 is set to green, to display heatbugs

  [colormap setColor: 64 ToName: "green"];

  // Colour 65 is set to white, used in this case below on probed heatbug.

  [colormap setColor: 65 ToName: "white"];

  for(i=66; i < 166; i++)
    [colormap setColor: i  ToRed: 0  Green: (double)(i-66) / 100.0 Blue: 0.4];


  [colormap setColor: 255 ToRed: 1 Green: 1 Blue: 0];
  // Now go in to the heatbugs in the model and set their colours to green (64)

  [[modelSwarm getCitizenList] forEach: M(setColor:) : (void *) 64];
  
  // Next, create a 2d window for display, set its size, zoom factor, title.
  

  rasterList = [List create: self];
  featureDisplayList = [List create: self];
  for(i=0; i<[Citizen getNumFeatures] ; i++)
 {
  char rasterName[20];

  sprintf(rasterName,"feature%02d",i);
    worldRaster = [ZoomRaster createBegin: self];
    [worldRaster setWindowTitle: strdup (rasterName)];
    [worldRaster setWindowGeometryRecordName: strdup (rasterName) ];
    //[worldRaster setWindowTitle: rasterName];
    //[worldRaster setWindowGeometryRecordName: rasterName ];
    worldRaster = [worldRaster createEnd];
    [worldRaster setColormap: colormap];
    [worldRaster setZoomFactor: 4];

    [worldRaster enableDestroyNotification: self
               notificationMethod: @selector (_worldRasterDeath_:)];
   
     [worldRaster setWidth: [[modelSwarm getPositionGrid] getSizeX]
	       Height: [[modelSwarm getPositionGrid] getSizeY]];
     [worldRaster pack];				  // draw the window.
     [rasterList addLast: worldRaster];

  // Now create a Value2dDisplay: this is a special object that will
  // display arbitrary 2d value arrays on a given Raster widget.
  
     featureDisplay = [Object2dDisplay createBegin: self];
     [featureDisplay setDisplayWidget: worldRaster];
     [featureDisplay setDiscrete2dToDisplay: [modelSwarm getPositionGrid]];
     [featureDisplay setObjectCollection: [modelSwarm getCitizenList] ];
     //  [featureDisplay setDisplayMessage: M(drawSelfOn:)];   // draw method
     [featureDisplay setDisplayMessage: M(drawSelfOn:)];

     featureDisplay = [featureDisplay createEnd];
     [featureDisplayList addLast: featureDisplay];

}

  
  // Also, tell the world raster to send mouse clicks to the heatbugDisplay
  // this allows the user to right-click on the display to probe the bugs.

  for(i=0; i<[Citizen getNumFeatures]; i++)
    {
  [worldRaster setButton: ButtonRight
	     /*  Client: protestDisplay*/ Client: [modelSwarm getPositionGrid]
               Message: M(makeProbeAtX:Y:)];
    }
  // Create the graph widget to display unhappiness.

   summaryGraph = [EZGraph createBegin: self];
   SET_WINDOW_GEOMETRY_RECORD_NAME (opinionGraph);
   [summaryGraph setTitle: "Opinion Changes Over Time"];
  [summaryGraph setAxisLabelsX: "time" Y: "summary information"];
  summaryGraph = [summaryGraph createEnd];
  [summaryGraph enableDestroyNotification: self
                notificationMethod: @selector (_opinionGraphDeath_:)];

  [summaryGraph createSequence: "Changes of Opinion"
                         withFeedFrom: modelSwarm 
                          andSelector: M(getNumberOfChangesThisStep)];
  [summaryGraph createSequence: "Interactions"
                         withFeedFrom: modelSwarm 
                          andSelector: M(getNumberOfInteractionsThisStep)];

   entropyGraph = [EZGraph createBegin: self];
   SET_WINDOW_GEOMETRY_RECORD_NAME (entropyGraph);
   [entropyGraph setTitle: "Perceived Similarity"];
   [entropyGraph setAxisLabelsX: "time" Y: "Average Proportion"];
    entropyGraph = [entropyGraph createEnd];


 [entropyGraph createSequence: "Perceived Harmony"
                        withFeedFrom: modelSwarm
                         andSelector: M(getHarmonyAvg)];
  

  [entropyGraph createSequence: "Perceived Acquaintance"
                         withFeedFrom: modelSwarm 
                          andSelector: M(getAcquaintanceAvg)];

  [entropyGraph createSequence: "Perceived Identical"
                         withFeedFrom: modelSwarm
                          andSelector: M(getIdenticalAvg)];


   [entropyGraph createSequence: "Total Entropy"
                         withFeedFrom: modelSwarm
                          andSelector: M(getMultidimensionalEntropy)];

  opinionGraph =  [NewEZGraph createBegin: self];
   SET_WINDOW_GEOMETRY_RECORD_NAME (opinionGraph);
   [opinionGraph setTitle: "Average Opinion"];
   [opinionGraph setAxisLabelsX: "time" Y: "Average Of Opinions"];
   opinionGraph = [opinionGraph createEnd];

  for(i=0; i < [Citizen getNumFeatures]; i++)
    {
      char  name[15];
      sprintf(name,"feature %d",i);
      [opinionGraph createSequence: name
                         withFeedFrom: modelSwarm  
                         andSelector: M(getAverageFeature:):(id)(long int)i];
    }
  return self;
}  

- _update_
{
  if (worldRaster)
    {
      id index;
      id rasterObject;

      //      [featureDisplayList forEach: M( display)];

      index=[rasterList begin: self];       
      while(( rasterObject = [index next]))
       [self drawWorld: rasterObject];
      [index drop];

#ifdef SCREENSHOTS
      {
        char filename[40];
        if(getCurrentTime() % 20==0)
	  {
        sprintf (filename, "%02dgraph%07d.png", [arguments getRunArg],getCurrentTime ());
        [actionCache doTkEvents];
        [[[[[[Pixmap createBegin: self]
              setWidget: [opinionGraph getGraph]]
             setDecorationsFlag: NO]
            createEnd] save: filename] drop];
        sprintf (filename, "raster%07d.png", getCurrentTime ());
        [[[[[[Pixmap createBegin: self]
              setWidget: worldRaster]
             setDecorationsFlag: YES]
            createEnd] save: filename] drop];
	  }
      }
      
#endif
    }
  if (opinionGraph)
    [opinionGraph step];


  if (entropyGraph)
    [entropyGraph step];
  if (summaryGraph)
    [summaryGraph step];  

  return self;
}

// Create the actions necessary for the simulation. This is where
// the schedule is built (but not run!)
// Here we create a display schedule - this is used to display the
// state of the world and check for user input. This schedule should
// be thought of as independent from the model - in particular, you
// will also want to run the model without any display.

- buildActions
{
  [super buildActions];
  
  // First, let our model swarm build its own schedule.

  [modelSwarm buildActions];
  
  // Create an ActionGroup for display: a bunch of things that occur in
  // a specific order, but at one step of simulation time. Some of these
  // actions could be executed in parallel, but we don't explicitly
  // notate that here.
  
  displayActions = [ActionGroup create: self];

  // Schedule up the methods to draw the display of the world
  // Also schedule the update of the unhappiness graph
  //  [displayActions createActionTo: unhappinessAverager message: M(update)];

  [displayActions createActionTo: self message: M(_update_)];

  // Schedule the update of the probe displays

  [displayActions createActionTo: probeDisplayManager message: M(update)];

  // Finally, schedule an update for the whole user interface code.
  // This is crucial: without this, no graphics update and the control
  // panel will be dead. It's best to put it at the end of the display schedule

  [displayActions createActionTo: actionCache message: M(doTkEvents)];

  // And the display schedule. Note the repeat interval is set from our
  // own Swarm data structure. Display is frequently the slowest part of a
  // simulation, so redrawing less frequently can be a help.

  displaySchedule = [Schedule createBegin: self];
  [displaySchedule setRepeatInterval: displayFrequency]; // note frequency!
  displaySchedule = [displaySchedule createEnd];
  [displaySchedule at: 0 createAction: displayActions];

  stopSchedule = [Schedule create: [self getZone]];
  [stopSchedule at: 1200
    createActionTo: self
         message: M(stopRunning)];


  
  return self;
}  

// activateIn: - activate the schedules so they're ready to run.
// The swarmContext argument has to do with what we were activated *in*.
// Typically the ObserverSwarm is the top-level Swarm, so it's activated
// in "nil". But other Swarms and Schedules and such will be activated
// inside of us.

- activateIn:  swarmContext
{
  // First, activate ourselves (just pass along the context).

  [super activateIn: swarmContext];

  // Activate the model swarm in ourselves. The model swarm is a
  // subswarm of the observer swarm.

  [modelSwarm activateIn: self];

  // Now activate our schedule in ourselves. This arranges for the
  // execution of the schedule we built.

  [displaySchedule activateIn: self];
 
  [stopSchedule activateIn: self]; 
  // Activate returns the swarm activity - the thing that's ready to run.

  return [self getActivity];
}


-stopRunning

{
  
  //    [[modelSwarm getActivity] terminate];
  
  //    [[self getActivity] terminate];
      [controlPanel setStateStopped]; 
	return self;
}


- drawFeature: (int) f PointX: (int) posX Y: (int) posY Color: (int) type {
//  id actor, neigh;

  
  id drawingRaster= [rasterList atOffset: f];
//  [drawingRaster fillRectangleX0: posX
//                Y0: posY
//                X1: posX+1
//                Y1: posY+1
//                Color: type];

  [drawingRaster drawPointX: posX Y: posY Color: type];
   return self;
}

- drawWorld: r
{
   [r drawSelf];
   // tkobjc_raster_flush (r);

   return self; 
}


// You could override the "go" method here if you want something special
// to happen when the model and observer actually start running. But
// the default GUISwarm go is probably good enough.

// - graphBug: aBug
// {
//   if (opinionGraph)
//     [opinionGraph createSequence: "Bug" 
//                   withFeedFrom: aBug 
//                   andSelector: M(getUnhappiness)];
//   return self;
// }


//Riolo's drawWorld method
////////////////////////////////////////////////////////////////////////
// drawWorld
// draw lines between agent's squares, and optionally, depending on
// the value of agentDraw, the values of the different features.
// That is, agentDraw == 0 means don't draw agent feature colors,
// but agentDraw in [1,numFeatures] means draw a color for
// the agentDraw-1 feature.  
// Note we have colors for only 15 features.
// Rick Riolo's draw world
// -drawWorld {
//   int x, y, worldSize = getUPar("worldSize");
//   int numFeatures = getUPar("numFeatures");
//   id r = worldRaster;
//   id world = [modelSwarm getWorld];

//   [r erase];

//   // draw most of the cells right and bottom borders
//   // but stop before we get to the rightmost and bottom-most.
//   for (y = 0; y < worldSize - 1; y++) {
//     for (x = 0; x < worldSize - 1; x++) {
//       Agent *c, *right, *down;
//       int rightSimilarity, downSimilarity;
// 	  char buff1[256], buff2[256];

//       c = [world getObjectAtX: x Y: y];
//       right = [world getObjectAtX: x+1 Y: y];
//       rightSimilarity = [c similarityTo: right];
//       down = [world getObjectAtX: x Y: y+1];
//       downSimilarity = [c similarityTo: down];

// 	  //fprintf( stderr, " agent %d (%d,%d,%s) rightSimilarity %d to %d (%d,%d,%s)\n",
// 	  //[c getID], [c getX], [c getY], [c getTraitValuesString: buff1], rightSimilarity,
// 	  //[right getID], [right getX], [right getY], [right getTraitValuesString: buff2] );


// 	  // note each agent is drawn over 2x2 cells.
// 	  if ( drawAgents > 0 ) {
// 		  [r drawPointX: x*3   Y: y*3   
// 			 Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		  [r drawPointX: x*3+1 Y: y*3   
// 			 Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		  [r drawPointX: x*3   Y: y*3+1 
// 			 Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		  [r drawPointX: x*3+1 Y: y*3+1 
// 			 Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 	  }
      
//       [r drawPointX: x*3+2 Y: y*3   Color: rightSimilarity];
//       [r drawPointX: x*3+2 Y: y*3+1 Color: rightSimilarity];
//       [r drawPointX: x*3   Y: y*3+2 Color: downSimilarity];
//       [r drawPointX: x*3+1 Y: y*3+2 Color: downSimilarity];
//     }
//   }

//   // do that bottom row, right only
//   y = worldSize - 1;
//   for (x = 0; x < worldSize - 1; x++) {
//     Agent *right, *c;

//     int rightSimilarity;
//     c = [world getObjectAtX: x Y: y];
//     right = [world getObjectAtX: x+1 Y: y];
//     rightSimilarity = [c similarityTo: right];
//     [r drawPointX: x*3+2 Y: y*3   Color: rightSimilarity];
//     [r drawPointX: x*3+2 Y: y*3+1 Color: rightSimilarity];

// 	if ( drawAgents > 0 ) {
// 		[r drawPointX: x*3   Y: y*3   
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		[r drawPointX: x*3+1 Y: y*3   
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		[r drawPointX: x*3   Y: y*3+1
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		[r drawPointX: x*3+1 Y: y*3+1 
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 	}
//   }

//   // do the right column down only
//   x = worldSize - 1;
//   for (y = 0; y < worldSize - 1; y++) {
//     Agent *down, *c;

//     int downSimilarity;
//     c = [world getObjectAtX: x Y: y];
//     down = [world getObjectAtX: x Y: y+1];
//     downSimilarity = [c similarityTo: down];
//     [r drawPointX: x*3   Y: y*3+2 Color: downSimilarity];
//     [r drawPointX: x*3+1 Y: y*3+2 Color: downSimilarity];

// 	if ( drawAgents > 0 ) {
// 		[r drawPointX: x*3   Y: y*3  
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		[r drawPointX: x*3+1 Y: y*3   
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		[r drawPointX: x*3   Y: y*3+1 
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 		[r drawPointX: x*3+1 Y: y*3+1 
// 		   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
// 	}
//   }

//   // and finally that bottom right agent.
//   if ( drawAgents > 0 ) {
//     Agent * c;
//     x = worldSize - 1;
//     y = worldSize - 1;
//     c = [world getObjectAtX: x Y: y];  
//     [r drawPointX: x*3   Y: y*3   
// 	   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
//     [r drawPointX: x*3+1 Y: y*3   
// 	   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
//     [r drawPointX: x*3   Y: y*3+1 
// 	   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
//     [r drawPointX: x*3+1 Y: y*3+1 
// 	   Color: [c getTraitValues][drawAgents-1] + numFeatures + 1];
//   }

//   [r drawSelf];

//   return self;
// }

@end
